from __future__ import annotations
from dataclasses import dataclass, field
from typing import Any, Dict, Iterable, FrozenSet, Optional, Set, Tuple

from .exceptions import DiffException, MergeException
from .utils import indent, color_green, color_yellow


Params = Tuple[str, ...]


@dataclass
class OptionDiff:
    name: Optional[str]
    added_params: Set[Params] = field(default_factory=set)
    removed_params: Set[Params] = field(default_factory=set)

    def __named_option_diff_str(self) -> str:
        string = f" {self.name}("

        if len(self.removed_params) == 0 and len(self.added_params) == 0:
            return string + ")"

        for removed_params in sorted(self.removed_params):
            string += f"\n-{indent(' '.join(removed_params))}"

        for added_params in sorted(self.added_params):
            string += f"\n+{indent(' '.join(added_params))}"

        return string + "\n )"

    def __positional_option_diff_str(self) -> str:
        string = ""

        if len(self.removed_params) == 0 and len(self.added_params) == 0:
            return string

        for removed_params in sorted(self.removed_params):
            string += f"-{' '.join(removed_params)}\n"

        for added_params in sorted(self.added_params):
            string += f"+{' '.join(added_params)}\n"

        string = string[:-1]

        return string

    def __str__(self) -> str:
        if self.name is not None:
            return self.__named_option_diff_str()

        return self.__positional_option_diff_str()


class Option:
    def __init__(self, name: Optional[str] = None, params: Optional[Iterable[Params]] = None) -> None:
        self.__name = name
        self.__params: Set[Params] = set(tuple(elem) for elem in params or [])

    @property
    def name(self) -> Optional[str]:
        return self.__name

    @property
    def params(self) -> FrozenSet[Params]:
        return frozenset(self.__params)

    def copy(self) -> Option:
        clone = Option(self.name)
        clone.merge(self)

        return clone

    def merge(self, other: Option) -> None:
        if self.name != other.name:
            raise MergeException(f"Cannot merge Options with different names: '{self.name}' and '{other.name}'")

        self.__params |= other.__params

    def diff(self, compared_to: Option) -> OptionDiff:
        diff = OptionDiff(self.name)

        if self.name != compared_to.name:
            raise DiffException(
                f"Cannot check differences of Options with different names: '{self.name}' and '{compared_to.name}'"
            )

        for their_params in compared_to.params:
            if their_params not in self.params:
                diff.removed_params.add(their_params)

        for our_params in self.params:
            if our_params not in compared_to.params:
                diff.added_params.add(our_params)

        return diff

    @staticmethod
    def from_dict(as_dict: Dict[str, Any]) -> Option:
        return Option(as_dict["name"], as_dict["params"])

    def to_dict(self) -> Dict[str, Any]:
        return {"name": self.__name, "params": tuple(self.__params)}

    def __repr__(self) -> str:
        return f"Option({repr(self.name)}, {repr(self.params)})"

    def __named_option_str(self, colored: bool = False) -> str:
        assert self.__name is not None

        string = f"{color_green(self.__name) if colored else self.__name}("

        if len(self.__params) == 0:
            return string + ")"

        if len(self.__params) == 1:
            params_str = " ".join(next(iter(self.__params)))
            return string + ((color_yellow(params_str) if colored else params_str) + ")")

        for params in sorted(self.__params):
            params_str = " ".join(params)
            string += f"\n{indent(color_yellow(params_str) if colored else params_str)}"
        string += "\n)"

        return string

    def __positional_option_str(self, colored: bool = False) -> str:
        string = ""

        for params in sorted(self.params):
            params_str = " ".join(params)
            string += (color_yellow(params_str) if colored else params_str) + "\n"

        string = string[:-1]

        return string

    def __str__(self) -> str:
        if self.name is not None:
            return self.__named_option_str()

        return self.__positional_option_str()

    def colored_str(self) -> str:
        if self.name is not None:
            return self.__named_option_str(colored=True)

        return self.__positional_option_str(colored=True)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Option):
            return False

        return self.__name == other.__name and self.__params == other.__params



from argparse import ArgumentParser, Namespace
from pathlib import Path
from typing import Optional

from syslog_ng_cfg_helper.driver_db import DriverDB, Driver
from syslog_ng_cfg_helper.driver_db.utils import color_red, unindent


def colorize_context_name(name: str, colored: bool = True) -> str:
    return color_red(name) if colored else name


def parse_args() -> Namespace:
    parser = ArgumentParser()
    parser.add_argument("--context", "-c", type=str, help="e.g.: destination")
    parser.add_argument("--driver", "-d", type=str, help="e.g.: http")
    parser.add_argument("--no-color", "-n", action="store_true", help="Do not color the output")

    return parser.parse_args()


def open_db() -> DriverDB:
    db_file = Path(__file__).parent / "syslog-ng-cfg-helper.db"
    with db_file.open("r") as file:
        driver_db = DriverDB.load(file)

    return driver_db


def print_global_options(driver_db: DriverDB, colored: bool) -> None:
    driver = driver_db.get_driver("options", DriverDB.GLOBAL_OPTIONS_DRIVER_NAME)
    global_options_str = driver.colored_str() if colored else str(driver)
    global_options_str = unindent(global_options_str)
    global_options_str_lines = global_options_str.split("\n")

    print("\n".join(global_options_str_lines[1:-1]))


def print_options(driver_db: DriverDB, context_name: str, driver_name: str, colored: bool) -> None:
    if context_name not in driver_db.contexts:
        print(f"The context '{context_name}' is not in the database.")
        print_contexts(driver_db, colored)
        return

    try:
        driver = driver_db.get_driver(context_name, driver_name)
        print(driver.colored_str() if colored else str(driver))
    except KeyError:
        print(
            f"The driver '{Driver.colorize_name(driver_name, colored)}' is not in the drivers of context "
            f"'{colorize_context_name(context_name, colored)}'."
        )
        print_drivers(driver_db, context_name, colored)


def print_drivers(driver_db: DriverDB, context_name: str, colored: bool) -> None:
    if context_name not in driver_db.contexts:
        print(f"The context '{colorize_context_name(context_name, colored)}' is not in the database.")
        print_contexts(driver_db, colored)
        return

    driver_names = sorted(driver.name for driver in driver_db.get_drivers_in_context(context_name))
    print(f"Drivers of context '{colorize_context_name(context_name, colored)}':")
    for driver_name in driver_names:
        print(f"  {Driver.colorize_name(driver_name, colored)}")
    print(
        f"Print the options of {Driver.colorize_name('DRIVER', colored)} with "
        f"`--context {context_name} --driver DRIVER`."
    )


def print_contexts(driver_db: DriverDB, colored: bool) -> None:
    print("Valid contexts:")
    for context_name in sorted(driver_db.contexts):
        print(f"  {colorize_context_name(context_name, colored)}")
    print(f"Print the drivers of {colorize_context_name('CONTEXT', colored)} with `--context CONTEXT`.")


def query(driver_db: DriverDB, context: Optional[str], driver: Optional[str], colored: bool) -> None:
    if context == "options":
        print_global_options(driver_db, colored)
        return

    if context and driver:
        print_options(driver_db, context, driver, colored)
        return

    if context and not driver:
        print_drivers(driver_db, context, colored)
        return

    if not context and driver:
        print(f"Please define the context of '{Driver.colorize_name(driver, colored)}' with `--context CONTEXT`.")
        return

    if not context and not driver:
        print_contexts(driver_db, colored)
        return


def run():
    args = parse_args()
    driver_db = open_db()
    query(driver_db, args.context, args.driver, not args.no_color)



from __future__ import annotations
from dataclasses import dataclass, field
from typing import Any, Dict, Optional, ValuesView

from .exceptions import DiffException, MergeException
from .option import Option, OptionDiff
from .utils import color_purple, diff_indent, indent, prepend_each_line, sorted_with_none


@dataclass
class BlockDiff:
    name: Optional[str]

    added_blocks: Dict[Optional[str], Block] = field(default_factory=dict)
    removed_blocks: Dict[Optional[str], Block] = field(default_factory=dict)
    changed_blocks: Dict[Optional[str], BlockDiff] = field(default_factory=dict)

    added_options: Dict[Optional[str], Option] = field(default_factory=dict)
    removed_options: Dict[Optional[str], Option] = field(default_factory=dict)
    changed_options: Dict[Optional[str], OptionDiff] = field(default_factory=dict)

    def __str__(self) -> str:
        string = f" {self.name}(\n"

        strs: Dict[Optional[str], str] = {}

        for block_name, block in self.added_blocks.items():
            strs[block_name] = f"{prepend_each_line(indent(str(block)), '+')}\n"

        for block_name, block in self.removed_blocks.items():
            strs[block_name] = f"{prepend_each_line(indent(str(block)), '-')}\n"

        for block_name, block_diff in self.changed_blocks.items():
            strs[block_name] = f"{diff_indent(str(block_diff))}\n"

        for option_name, option in self.added_options.items():
            strs[option_name] = f"{prepend_each_line(indent(str(option)), '+')}\n"

        for option_name, option in self.removed_options.items():
            strs[option_name] = f"{prepend_each_line(indent(str(option)), '-')}\n"

        for option_name, option_diff in self.changed_options.items():
            strs[option_name] = f"{diff_indent(str(option_diff))}\n"

        for block_or_option_name in sorted_with_none(strs.keys()):
            string += strs[block_or_option_name]

        string += " )"

        return string


class Block:
    def __init__(self, name: str):
        self.__name = name
        self.__blocks: Dict[str, Block] = {}
        self.__options: Dict[Optional[str], Option] = {}

    @property
    def name(self) -> str:
        return self.__name

    @property
    def blocks(self) -> ValuesView[Block]:
        return self.__blocks.values()

    def get_block(self, name: str) -> Block:
        return self.__blocks[name]

    def add_block(self, block: Block) -> None:
        if block.name not in self.__blocks:
            self.__blocks[block.name] = block.copy()
        else:
            self.get_block(block.name).merge(block)

    def remove_block(self, name) -> None:
        self.__blocks.pop(name)

    @property
    def options(self) -> ValuesView[Option]:
        return self.__options.values()

    def get_option(self, name: Optional[str]) -> Option:
        return self.__options[name]

    def add_option(self, option: Option) -> None:
        if option.name not in self.__options:
            self.__options[option.name] = option.copy()
        else:
            self.get_option(option.name).merge(option)

    def remove_option(self, name: Optional[str]) -> None:
        self.__options.pop(name)

    def merge(self, other: Block) -> None:
        if self.name != other.name:
            raise MergeException(f"Cannot merge two Blocks with different names: '{self.name}' and '{other.name}'")

        for block in other.blocks:
            self.add_block(block)

        for option in other.options:
            self.add_option(option)

    def copy(self) -> Block:
        clone = Block(self.name)
        clone.merge(self)

        return clone

    def __process_option_to_block_transform_diff(
        self,
        their_option_name: str,
        their_option: Option,
        diff: BlockDiff,
    ) -> None:
        our_block = self.get_block(their_option_name)

        their_block_from_option = Block(their_option_name)
        their_block_from_option.add_option(Option(params=their_option.params))

        diff.changed_blocks[their_option_name] = our_block.diff(their_block_from_option)

    def __gather_option_diffs(self, compared_to: Block, diff: BlockDiff) -> None:
        for their_option_name, their_option in compared_to.__options.items():
            if their_option_name not in self.__options:
                if their_option_name not in self.__blocks:
                    diff.removed_options[their_option_name] = their_option.copy()
                else:
                    self.__process_option_to_block_transform_diff(their_option_name, their_option, diff)
                continue

            our_option = self.get_option(their_option_name)
            if our_option == their_option:
                continue

            diff.changed_options[their_option_name] = our_option.diff(their_option)

        for our_option_name, our_option in self.__options.items():
            if our_option_name not in compared_to.__options:
                diff.added_options[our_option_name] = our_option.copy()
                continue

    def __process_block_to_option_transform_diff(
        self,
        their_block_name: str,
        their_block: Block,
        diff: BlockDiff,
    ) -> None:
        our_option = self.get_option(their_block_name)

        our_block_from_option = Block(their_block_name)
        our_block_from_option.add_option(Option(params=our_option.params))

        diff.changed_blocks[their_block_name] = our_block_from_option.diff(their_block)

    def __gather_block_diffs(self, compared_to: Block, diff: BlockDiff) -> None:
        for their_block_name, their_block in compared_to.__blocks.items():
            if their_block_name not in self.__blocks:
                if their_block_name not in self.__options:
                    diff.removed_blocks[their_block_name] = their_block.copy()
                else:
                    self.__process_block_to_option_transform_diff(their_block_name, their_block, diff)
                continue

            our_block = self.get_block(their_block_name)
            if our_block == their_block:
                continue

            diff.changed_blocks[their_block_name] = our_block.diff(their_block)

        for our_block_name, our_block in self.__blocks.items():
            if our_block_name not in compared_to.__blocks:
                diff.added_blocks[our_block_name] = our_block.copy()
                continue

    def diff(self, compared_to: Block) -> BlockDiff:
        diff = BlockDiff(self.name)

        if self.name != compared_to.name:
            raise DiffException(
                f"Cannot check differences of Blocks with different names: '{self.name}' and '{compared_to.name}'"
            )

        self.__gather_option_diffs(compared_to, diff)
        self.__gather_block_diffs(compared_to, diff)

        return diff

    @staticmethod
    def from_dict(as_dict: Dict[str, Any]) -> Block:
        self = Block(as_dict["name"])

        for block in as_dict["blocks"].values():
            self.add_block(Block.from_dict(block))

        for option in as_dict["options"].values():
            self.add_option(Option.from_dict(option))

        return self

    def to_dict(self) -> Dict[str, Any]:
        as_dict: Dict[str, Any] = {
            "name": self.name,
            "blocks": {},
            "options": {},
        }

        for block_name, block in self.__blocks.items():
            as_dict["blocks"][block_name] = block.to_dict()

        for option_name, option in self.__options.items():
            as_dict["options"][option_name or ""] = option.to_dict()

        return as_dict

    def __repr__(self) -> str:
        return f"Block({repr(self.__name)}, {repr(self.__blocks)}, {repr(self.__options)})"

    def __str(self, colored: bool = False) -> str:
        string = f"{self.colorize_name(self.name, colored)}(\n"

        block_and_option_strs: Dict[Optional[str], str] = {}

        for block_name, block in self.__blocks.items():
            block_and_option_strs.setdefault(block_name, "")
            block_and_option_strs[block_name] += f"{indent(block.colored_str() if colored else str(block))}\n"

        for option_name, option in self.__options.items():
            block_and_option_strs.setdefault(option_name, "")
            block_and_option_strs[option_name] += f"{indent(option.colored_str() if colored else str(option))}\n"

        for block_or_option_name in sorted_with_none(block_and_option_strs.keys()):
            string += block_and_option_strs[block_or_option_name]

        string += ")"

        return string

    def __str__(self) -> str:
        return self.__str()

    def colored_str(self) -> str:
        return self.__str(colored=True)

    @staticmethod
    def colorize_name(name: str, colored: bool = True) -> str:
        return color_purple(name) if colored else name

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Block):
            return False

        return self.__name == other.__name and self.__blocks == other.__blocks and self.__options == other.__options



import sys

from argparse import ArgumentParser, Namespace
from pathlib import Path

from syslog_ng_cfg_helper.driver_db import DriverDB


def parse_args() -> Namespace:
    parser = ArgumentParser()
    parser.add_argument(
        "--old-db-file",
        "-o",
        type=str,
        required=True,
        help="Path of the old syslog-ng-cfg-helper.db file.",
    )
    parser.add_argument(
        "--new-db-file",
        "-n",
        type=str,
        required=True,
        help="Path of the new syslog-ng-cfg-helper.db file.",
    )

    return parser.parse_args()


def main() -> int:
    args = parse_args()

    old_db_file = Path(args.old_db_file)
    new_db_file = Path(args.new_db_file)

    with old_db_file.open("r", encoding="utf-8") as file:
        old_driver_db = DriverDB.load(file)

    with new_db_file.open("r", encoding="utf-8") as file:
        new_driver_db = DriverDB.load(file)

    print(new_driver_db.diff(old_driver_db))

    return 0


if __name__ == "__main__":
    sys.exit(main())
